#include "amxxmodule.h"
#include <curl\curl.h>
#include "CurlThread.h"
#include "amx_utils.h"
#include "curl_utils.h"

CCurlThread::CCurlThread(AMX* plugin, CURL* handle) {
	m_pAMXPlugin = plugin;
	m_pCURLHandle = handle;

	m_amxCompliteCallback = 0;
	m_amxCompliteCbackData = NULL;
	m_amxCompliteCbackDataSize = 0;

	m_bMustExit = false;
}

CCurlThread::~CCurlThread() {
	if(m_amxCompliteCbackData) {
		delete[m_amxCompliteCbackDataSize] m_amxCompliteCbackData;
	}
}

void CCurlThread::SetAmxCompliteCallback(char* function, cell* data, int data_len) {
	if(m_amxCompliteCallback) {
		MF_UnregisterSPForward(m_amxCompliteCallback);
	}

	if(m_amxCompliteCbackData) {
		delete[m_amxCompliteCbackDataSize] m_amxCompliteCbackData;
	}

	if(strlen(function) == 0) {
		return;
	}

	m_amxCompliteCbackData = data;
	m_amxCompliteCbackDataSize = data_len;
	m_amxCompliteCallback = MF_RegisterSPForwardByName(m_pAMXPlugin, function, FP_CELL /* CURLcode */, FP_CELL /*CURL* */, FP_ARRAY /* data */, FP_CELL /* data len */, FP_DONE);
	
	if(m_amxCompliteCallback < 0) {
		MF_LogError(m_pAMXPlugin, AMX_ERR_NATIVE, "(CCurlThread::SetAmxCompliteCallback) Callback function \"%s\" not found.", function);
	}
}

void CCurlThread::SetAmxCallback(CURLoption opt, char* function) {
	if(m_amxCallbacks.count(opt)) {
		MF_UnregisterSPForward(m_amxCallbacks[opt]);
	}

	char params[8];
	int count = GetAmxCallbackParams(opt, params);

	if(!count) {
		MF_LogError(m_pAMXPlugin, AMX_ERR_NATIVE, "(CCurlThread::SetAmxCallback) Unknown curl option \"%d\".", (int)opt);
		return;
	}

	int callback = MF_RegisterSPForwardByName_arr(m_pAMXPlugin, function, params, count);
	if(callback < 0) {
		m_amxCallbacks.erase(opt);

		MF_LogError(m_pAMXPlugin, AMX_ERR_NATIVE, "(CCurlThread::SetAmxCallback) Callback function \"%s\" not found.", function);
		return;
	}

	m_amxCallbacks[opt] = callback;
}

int CCurlThread::GetAmxCallback(CURLoption opt) {
	if(m_amxCallbacks.count(opt)) {
		return m_amxCallbacks[opt];
	}

	return -1;
}

void CCurlThread::SetAmxCallbackData(CURLoption opt, cell data) {
	if(data == 0) {
		m_amxCBacksData.erase(opt);
		return;
	}

	m_amxCBacksData[opt] = data;
}

cell CCurlThread::GetAmxCallbackData(CURLoption opt) {
	if(!m_amxCBacksData.count(opt)) {
		return 0;
	}
	
	return m_amxCBacksData[opt];
}

void CCurlThread::ExecuteNextAmxCallback() {
	m_locker.Lock();

	if(m_executeCBackData.callback > -1) {
		cell params[6]; //  6       
		for(int i = 0; i < m_executeCBackData.count; i++) {
			TAmxCallbackParam* p = &m_executeCBackData.params[i];
			
			switch(p->type) {
			case FP_CELL:
				params[i] = p->iData;
				break;
			case FP_FLOAT:
				params[i] = amx_ftoc(p->fData);
				break;
			case FP_STRING:
			case FP_STRINGEX:
				params[i] = (cell)p->pData;
				break;
			case FP_ARRAY:
			case _FP_ARRAY_CELLS:
				params[i] = MF_PrepareCellArray((cell*)p->pData, p->size);
				break;
			case _FP_ARRAY_CELLSA:
				params[i] = MF_PrepareCellArrayA((cell*)p->pData, p->size, true);
				break;
			case _FP_ARRAY_CHARS:
				params[i] = MF_PrepareCharArray((char*)p->pData, p->size);
				break;
			case _FP_ARRAY_CHARSA:
				params[i] = MF_PrepareCharArrayA((char*)p->pData, p->size, true);
				break;
			default:
				MF_LogError(m_pAMXPlugin, AMX_ERR_NATIVE, "(CCurlThread::ExecuteNextAmxCallback) Unknown param type \"%d\".", p->type);
				params[i] = 0;
			}
		}

		m_executeCBackData.result.data = MF_ExecuteForward_arr(m_pAMXPlugin, m_executeCBackData.callback, params, m_executeCBackData.count);
		m_executeCBackData.result.ready = true;
		m_executeCBackData.callback = -1;
	}

	m_locker.Unlock();

	m_cvReady.WakeOne();
}

void CCurlThread::SafeExit() {
	m_locker.Lock();

	m_bMustExit = true;

	m_locker.Unlock();

	m_cvReady.WakeOne();

	Wait();
}

cell CCurlThread::ExecuteAmxCallback(CURLoption opt, TAmxCallbackParam* params, int count) {
	return ExecuteAmxCallback(m_amxCallbacks[opt], params, count, opt);
}

cell CCurlThread::ExecuteAmxCallback(int callback, TAmxCallbackParam* params, int count, CURLoption opt = CURLOPT_LASTENTRY) {
	m_locker.Lock();

	m_executeCBackData.callback = callback;
	m_executeCBackData.params = params;
	m_executeCBackData.count = count;

	while(!m_executeCBackData.result.ready && !m_bMustExit) {
		m_cvReady.Wait(m_locker);
	}

	m_executeCBackData.result.ready = false;

	if(m_bMustExit) {
		m_locker.Unlock();

		if(opt == CURLOPT_LASTENTRY) {
			return 0;
		}

		return CurlLinkedInstError(opt);
	}

	m_locker.Unlock();

	return m_executeCBackData.result.data;
}

//overload
void CCurlThread::Run() {
	CURLcode code = curl_easy_perform(m_pCURLHandle);

	if(m_amxCompliteCallback) {
		TAmxCallbackParam params[4];
		params[0].Set(code);
		params[1].Set((cell)m_pCURLHandle);
		params[2].Set(m_amxCompliteCbackData, _FP_ARRAY_CELLS, m_amxCompliteCbackDataSize);
		params[3].Set(m_amxCompliteCbackDataSize);
		
		ExecuteAmxCallback(m_amxCompliteCallback, params, 4);
	}

	m_locker.Lock();
	if(m_bMustExit) {
		m_locker.Unlock();

		ExecuteNextAmxCallback();
	}
	else {
		m_locker.Unlock();
	}

	Exit();
}